// MMSQ: 2048 32 1 256

__kernel void
multi_square_2048_32_1(
	__global uint *gx,
	__constant uint *m,
	uint m0i, int w)
{
	uint ga;
	uint x[64], d[64];

	/*
	 * Read value from global memory.
	 */
	ga = get_global_id(0) * 64;
	#pragma unroll 64
	for (int i = 0; i < 64; i ++) {
		x[i] = gx[ga + i];
	}

	for (int k = 0; k < w; k ++) {
		ulong dh;
		uint cc, r3;
		bool nsub;

		for (int i = 0; i < 64; i ++) {
			d[i] = 0;
		}
		dh = 0;
		#pragma unroll 1
		for (int i = 0; i < 64; i ++) {
			uint f, xi;
			ulong r1, r2, zh;

			xi = x[i];
			f = (d[0] + xi * x[0]) * m0i;
			r1 = 0;
			r2 = 0;
			#pragma unroll 64
			for (int j = 0; j < 64; j ++) {
				ulong z;
				uint t;

				z = (ulong)d[j] + (ulong)xi * (ulong)x[j] + r1;
				r1 = z >> 32;
				t = (uint)z;
				z = (ulong)t + (ulong)f * (ulong)m[j] + r2;
				r2 = z >> 32;
				if (j != 0) {
					d[j - 1] = (uint)z;
				}
			}
			zh = dh + r1 + r2;
			d[63] = (uint)zh;
			dh = zh >> 32;
		}

		/*
		 * Copy and possibly subtract the modulus to ensure that
		 * the value fits.
		 */
		nsub = dh != 0;
		cc = 0;
		#pragma unroll 64
		for (int i = 0; i < 64; i ++) {
			uint xi;

			xi = d[i];
			if (nsub) {
				uint yi;

				yi = xi - m[i] - cc;
				cc = (yi > xi) | (cc & (yi == xi));
				xi = yi;
			}
			x[i] = xi;
		}
	}

	/*
	 * Write back result into global memory.
	 */
	for (int i = 0; i < 64; i ++) {
		gx[ga + i] = x[i];
	}
}
